#include "../../include/error.h"
#include "../../include/pool.h"
#include "../../include/util.h"
#include "../../include/ctx.h"

#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>

bool lm_ctx_removeable(lm_ctx_t *ctx, lm_entry_t *entry){
  if(NULL == ctx && NULL == entry){
    lm_error_set(LM_ERR_ArgNULL);
    return false;
  }

  if(!lm_ctx_database_is_installed(ctx, entry->name, entry->version)){
    lm_error_set(LM_ERR_PkgNotInstalled);
    return false;
  }

  if(!lm_ctx_database_init(ctx))
    return false;

  lm_entry_t cur;

  while (lm_database_entry_next(ctx->db, &cur)) {
    if(lm_package_data_depends(&cur, entry->name)){
      lm_error_set(LM_ERR_PkgBreaks, cur.name);
      lm_database_entry_next_free(ctx->db, &cur);
      return false;
    }
  }

  lm_database_entry_next_free(ctx->db, &cur);
  return true;
}

bool lm_ctx_remove(lm_ctx_t *ctx, lm_entry_t *entry, lm_ctx_remove_callback_t callback, void *data){
  if(NULL == ctx && NULL == entry){
    lm_error_set(LM_ERR_ArgNULL);
    return false;
  }

  if(NULL == ctx->root){
    lm_error_set(LM_ERR_CtxRootNULL);
    return false;
  }

  if(!lm_ctx_database_init(ctx))
    return false;

  char *path = NULL, *hash = NULL, *fpath = NULL;
  bool in_keep = false, ret = false;
  size_t total = 0, current = 0;

  total = lm_database_files_count(ctx->db, entry);
  pdebug(__func__, "removing %lu files", total);

  while(lm_database_files_next(ctx->db, entry, &path, &hash, &in_keep)){
    if(in_keep){
      pdebug(__func__, "not removing file because it is set as keep: %s", path);
      goto next;
    }

    fpath = join_alloc(ctx->root, path);
    pdebug(__func__, "removing file %s (%s)", fpath, hash);

    if(!exists(fpath, NULL)){
      pdebug(__func__, "found file in database, but its not on the file system: %s", fpath);
      goto next;
    }
   
    if(unlink(fpath) < 0){
      pdebug(__func__, "failed to delete file for removing %s: %s", entry->name, strerror(errno));
      lm_error_set(LM_ERR_PkgFileUnlinkFail, path, strerror(errno));
      goto next;
    }

  next:
    free(fpath);
    fpath = NULL;

    if(NULL != callback && !callback(ctx, entry, fpath, ++current, total, data))
      goto end;
  }

  if(!lm_database_entry_del(ctx->db, entry)){
    char *suberr = lm_strerror_dup();
    pdebug(__func__, "failed to delete %s from the database: %s", entry->name, lm_strerror());
    lm_error_set(LM_ERR_PkgDatabaseDelFail, suberr);
    free(suberr);
    goto end;
  }

  if(!lm_database_files_del(ctx->db, entry)){
    char *suberr = lm_strerror_dup();
    pdebug(__func__, "failed to delete files of %s from the database: %s", entry->name, suberr);
    lm_error_set(LM_ERR_PkgFilesDelFail, suberr);
    free(suberr);
    goto end;
  }

  if(!lm_database_changes_del(ctx->db, entry)){
    char *suberr = lm_strerror_dup();
    pdebug(__func__, "failed to delete changes file for %s: %s", entry->name, suberr);
    lm_error_set(LM_ERR_PkgChangesDelFail, suberr);
    goto end;
  }

  ret = true;
end: 
  lm_database_files_next_free(ctx->db, entry, &path, &hash, &in_keep);
  return ret;
}

