#include "../../include/database.h"
#include "../../include/package.h"
#include "../../include/error.h"
#include "../../include/util.h"

#include <sqlite3.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>

void lm_entry_init(lm_entry_t *entry){
  bzero(entry, sizeof(lm_entry_t));
}

void lm_entry_free(lm_entry_t *entry){
  if(NULL == entry)
    return;
  lm_package_data_free(entry);
  lm_entry_init(entry);
}

bool lm_database_entry_add(lm_database_t *db, lm_entry_t *entry){
  if(NULL == db || NULL == entry){
    lm_error_set(LM_ERR_ArgNULL);
    return false;
  }

  char depends[lm_package_data_depend_strlen(entry)];
  bool ret = false;

  if(sqlite3_prepare(db->sql, queries[QUERY_INSERT_SINGLE], strlen(queries[QUERY_INSERT_SINGLE]), &db->st, NULL) != SQLITE_OK){
    pdebug(__func__, "failed to prepare statement for inserting %s: %s", entry->name, sqlite3_errmsg(db->sql));
    lm_error_set(LM_ERR_DbSqlPrepareFail);
    goto end;
  }

  sqlite3_bind_text(db->st, COLUMN_NAME, entry->name, strlen(entry->name), SQLITE_STATIC);
  sqlite3_bind_text(db->st, COLUMN_DESC, entry->desc, strlen(entry->desc), SQLITE_STATIC);
  sqlite3_bind_text(db->st, COLUMN_VERSION, entry->version, strlen(entry->version), SQLITE_STATIC);
  sqlite3_bind_int64(db->st, COLUMN_SIZE, entry->size);

  if(!lm_package_data_depend_tostr(entry, depends)){
    pdebug(__func__, "failed to convert depends to string for inserting %s: %s", entry->name, lm_strerror());
    goto end;
  }
  pdebug(__func__, "depend list for %s: %s", entry->name, depends);
  sqlite3_bind_text(db->st, COLUMN_DEPENDS, depends, strlen(depends), SQLITE_STATIC);

  if(!lm_database_step_all(db->st)){
    pdebug(__func__, "failed to execute insert statement for inserting %s: %s", entry->name, sqlite3_errmsg(db->sql));
    lm_error_set(LM_ERR_DbSqlInsertFail);
    goto end;
  }

  ret = true;
end:
  if(NULL != db->st){
    sqlite3_finalize(db->st);
    db->st = NULL;
  }
  return ret;
}

bool lm_database_entry_find(lm_database_t *db, lm_entry_t *entry, char *name, char *version){
  if(NULL == db || NULL == name){
    lm_error_set(LM_ERR_ArgNULL);
    return false;
  }

  bool ret = false;
  char *query = NULL;

  if(NULL == version)
    query = queries[QUERY_SELECT_SINGLE_1];
  else
    query = queries[QUERY_SELECT_SINGLE_2];

  // package pointer should already be free'd
  // we are initing it just in case the caller didn't
  if(NULL != entry)
    lm_entry_init(entry);

  if(sqlite3_prepare(db->sql, query, strlen(query), &db->st, NULL) != SQLITE_OK){
    pdebug(__func__, "failed to prepare statement for finding %s: %s", name, sqlite3_errmsg(db->sql));
    lm_error_set(LM_ERR_DbSqlPrepareFail);
    goto end;
  }

  sqlite3_bind_text(db->st, COLUMN_NAME, name, strlen(name), SQLITE_STATIC);
  if(NULL != version)
    sqlite3_bind_text(db->st, COLUMN_VERSION, version, strlen(version), SQLITE_STATIC);

  if(sqlite3_step(db->st) != SQLITE_ROW){
    pdebug(__func__, "got no rows for %s", name);
    lm_error_set(LM_ERR_DbSqlNotFound);
    goto end;
  }

  if(NULL == entry){
    ret = true;
    goto end;
  }

  entry->name = strdup((char*)sqlite3_column_text(db->st, COLUMN_NAME-1));
  entry->desc = strdup((char*)sqlite3_column_text(db->st, COLUMN_DESC-1));
  entry->version = strdup((char*)sqlite3_column_text(db->st, COLUMN_VERSION-1));
  entry->size = sqlite3_column_int64(db->st, COLUMN_SIZE-1);

  char *depends = (char*)sqlite3_column_text(db->st, COLUMN_DEPENDS-1);
  if(!lm_package_data_depend_fromstr(entry, depends)){
    pdebug(__func__, "failed to load depends for finding %s: %s", entry->name, lm_strerror());
    // error is set by the function
    goto end;
  }

  ret = true;
end:
  if(NULL != db->st){
    sqlite3_finalize(db->st);
    db->st = NULL;
  }
  return ret;
}

bool lm_database_entry_del(lm_database_t *db, lm_entry_t *entry){
  if(NULL == db || NULL == entry){
    lm_error_set(LM_ERR_ArgNULL);
    return false;
  }

  bool ret = false;

  if(sqlite3_prepare(db->sql, queries[QUERY_DELETE_SINGLE], strlen(queries[QUERY_DELETE_SINGLE]), &db->st, NULL) != SQLITE_OK){
    pdebug(__func__, "failed to prepare statement for deleting %s: %s", entry->name, sqlite3_errmsg(db->sql));
    lm_error_set(LM_ERR_DbSqlPrepareFail);
    goto end;
  }

  sqlite3_bind_text(db->st, COLUMN_NAME, entry->name, strlen(entry->name), SQLITE_STATIC);

  if(sqlite3_step(db->st) != SQLITE_DONE){
    pdebug(__func__, "failed to execute delete statement for deleting %s: %s", entry->name, sqlite3_errmsg(db->sql));
    lm_error_set(LM_ERR_DbSqlDeleteFail);
    goto end;
  }

  ret = true;
end:
  if(NULL != db->st){
    sqlite3_finalize(db->st);
    db->st = NULL;
  }
  return ret;
}

bool lm_database_entry_next(lm_database_t *db, lm_entry_t *entry){
  if(NULL == db || NULL == entry){
    lm_error_set(LM_ERR_ArgNULL);
    return false;
  }

  if(NULL != db->st)
    lm_entry_free(entry);

  else if(NULL == db->st &&
     sqlite3_prepare(db->sql, queries[QUERY_SELECT_ALL], strlen(queries[QUERY_SELECT_ALL]), &db->st, NULL) != SQLITE_OK){
    pdebug(__func__, "failed to prepare statement for selecting all: %s", sqlite3_errmsg(db->sql));
    lm_error_set(LM_ERR_DbSqlPrepareFail);
    return false;
  }

  lm_entry_init(entry);

  if(sqlite3_step(db->st) != SQLITE_ROW){
    sqlite3_finalize(db->st);
    db->st = NULL;
    return false;
  }

  entry->name = strdup((char*)sqlite3_column_text(db->st, COLUMN_NAME-1));
  entry->desc = strdup((char*)sqlite3_column_text(db->st, COLUMN_DESC-1));
  entry->version = strdup((char*)sqlite3_column_text(db->st, COLUMN_VERSION-1));
  entry->size = sqlite3_column_int64(db->st, COLUMN_SIZE-1);

  char *depends = (char*)sqlite3_column_text(db->st, COLUMN_DEPENDS-1);
  if(!lm_package_data_depend_fromstr(entry, depends)){
    pdebug(__func__, "failed to load depends for finding %s: %s", entry->name, lm_strerror());
    sqlite3_finalize(db->st);
    return false;
  }

  return true;
}

void lm_database_entry_next_free(lm_database_t *db, lm_entry_t *entry){
  if(NULL == db || NULL == entry){
    lm_error_set(LM_ERR_ArgNULL);
    return;
  }

  if(NULL != db->st){
    sqlite3_finalize(db->st);
    db->st = NULL;
  }

  if(NULL != entry)
    lm_entry_free(entry);
}
