#include "../../include/package.h"
#include "../../include/error.h"
#include "../../include/pool.h"
#include "../../include/mptp.h"
#include "../../include/util.h"

#include <stdbool.h>
#include <dirent.h>
#include <libgen.h>
#include <stdlib.h>
#include <string.h>

bool lm_pool_list_load(lm_pool_t *pool, char *dir){
  if(NULL == pool || NULL == dir){
    lm_error_set(LM_ERR_ArgNULL);
    return false;
  }
  
  lm_pool_list_free(pool);

  if(lm_pool_path_is_empty(pool)){
    lm_error_set(LM_ERR_PoolPathsEmpty);
    return false;
  }

  if(!exists(pool->list_file, NULL) || !can_read(pool->list_file)){
    lm_error_set(LM_ERR_PoolListCantRead);
    return false;
  }

  pdebug(__func__, "(%s) extracting pool to %s", pool->name, dir);

  if(!mkdir_ifnot(pool->dir, 0755)){
    lm_error_set(LM_ERR_PoolBadDir);
    return false;
  }

  if(!mkdir_ifnot(dir, 0755)){
    lm_error_set(LM_ERR_PoolListBadDir);
    return false;
  }

  size_t ent_len = 0, list_dir_len = strlen(dir);
  struct dirent *ent = NULL;
  DIR *dirfd = NULL;
  bool ret = false;

  if(!extract_archive(dir, pool->list_file))
    goto end;

  if((dirfd = opendir(dir)) == NULL){
    lm_error_set(LM_ERR_PoolListDirFail);
    goto end;
  }
  
  while((ent = readdir(dirfd)) != NULL){
    if(eq(ent->d_name, "..") || eq(ent->d_name, "."))
      continue;

    ent_len = strlen(ent->d_name);
    char datap[ent_len+list_dir_len+10];
    join_multiple(datap, dir, ent->d_name, "DATA");
    
    lm_pkg_t *pkg = lm_package_new();
    
    if(!lm_package_data_load(&pkg->data, datap)){
      pdebug(__func__, "(%s) failed to load new package from %s", pool->name, datap);
      lm_package_free(pkg);
      continue;
    }

    if(!lm_pool_package_add(pool, pkg)){
      pdebug(__func__, "(%s) failed to add new package: %s", pool->name, pkg->data.name);
      lm_package_free(pkg);
      continue;
    }
      
    pdebug(__func__, "(%s) added new package: %s", pool->name, pkg->data.name);
  }

  ret = true;

end:
  if(NULL != dirfd)
    closedir(dirfd);
  return ret;
}

bool lm_pool_list_download(lm_pool_t *pool, char *dir, lm_mptp_transfer_callback_t callback, void *data) {
  if(NULL == pool || NULL == dir){
    lm_error_set(LM_ERR_ArgNULL);
    return false;
  }

  if(!pool->available){
    lm_error_set(LM_ERR_PoolNotAvailable);
    return false;
  }

  if(lm_pool_path_is_empty(pool)){
    lm_error_set(LM_ERR_PoolPathsEmpty);
    return false;
  }

  if(!mkdir_ifnot(pool->dir, 0755)){
    lm_error_set(LM_ERR_PoolBadDir);
    return false;
  }

  if(NULL == pool->url.path || NULL == pool->url.host){
    lm_error_set(LM_ERR_PoolUrlEmpty);
    return false;
  }

  int sock = -1;
  lm_mptp_t packet;
  bool ret = false;

  if((sock = lm_mptp_client_connect(pool->url.host, pool->url.port)) < 0)
    return false;

  lm_mptp_new(&packet, true, MPTP_C2S_LIST, true);
  lm_mptp_set_host(&packet, pool->url.host);
  lm_mptp_set_path(&packet, pool->url.path);

  if(!lm_mptp_client_send(sock, &packet)){
    pdebug(__func__, "list file request failed for %s: %s", pool->name, lm_strerror());
    goto end;
  }
 
  if(!lm_mptp_recvfile(sock, pool->list_file, callback, data)){
    pdebug(__func__, "recvfile failed for %s: %s", pool->name, lm_strerror());
    goto end;
  }

  ret = true;
end:
  lm_mptp_free(&packet);
  lm_mptp_close(sock);
  if(ret)
    ret = lm_pool_list_load(pool, dir);
  return ret;
}

void lm_pool_list_free(lm_pool_t *pool){
  if(NULL == pool)
    return;

  lm_pkg_t *cur = pool->pkg, *prev = NULL;
  while(NULL != cur){
    prev = cur;
    cur = cur->next;

    lm_package_free(prev);
    free(prev);
  }
}
